import os
import re
import string

import dask.bag as db
import pandas as pd
import requests
from bs4 import BeautifulSoup

_NPARTITIONS = os.cpu_count()


def dask_bag_sequence_parallel(sequence, func, npartitions=_NPARTITIONS):
    return list(db.from_sequence(sequence, npartitions=npartitions).map(func))


_ENCODING = 'utf-8'
_PARSER = 'lxml'
_TIMEOUT = 3


headers = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 \
    (KHTML, like Gecko) Chrome/68.0.3440.106 Safari/537.36'
}


def BS(uri):
    if uri.startswith('http'):
        return BeautifulSoup(requests.get(uri, headers=headers,
                                          timeout=_TIMEOUT).content, _PARSER)
    else:
        return BeautifulSoup(open(uri, 'r', encoding='utf-8'), _PARSER)

# a


def concat_tag_a(a):
    if 'title' in a.attrs:
        return a.attrs['title']+'_'+a.attrs['href']
    else:
        return get_text(a)+'_'+a.attrs['href']


def dict_tag_a(a):
    if 'title' in a.attrs:
        return {'title': a.attrs['title'], 'href': a.attrs['href']}
    else:
        return {'title': get_text(a), 'href': a.attrs['href']}


def a_dict_list(ele):
    return [dict_tag_a(a) for a in ele.find_all('a')]

# text


def get_text(ele):
    return re.sub(' |\x0b|\x0c|\n|\r|\t|\s', '', ele.get_text().strip())


def get_text_by_class(ele, tag, class_):
    return get_text(find_by_class(ele, tag, class_))


def get_text_by_id(ele, tag, class_):
    return get_text(find_by_id(ele, tag, class_))


# find


def find_by_tag(ele, tag):
    return ele.find(tag)


def find_all_by_tag(ele, tag):
    return ele.find_all(tag)


def find_by_text(ele, text):
    return ele.find(text=re.compile(text))


def find_by_selector(ele, selector):
    return ele.select_one(selector)


def find_all_by_text(ele, text):
    return ele.find_all(text=re.compile(text))


def find_by_class(ele, tag, class_):
    return ele.find(tag, attrs={'class': class_})


def find_all_by_class(ele, tag, class_):
    return ele.find_all(tag, attrs={'class': class_})


def find_all_by_selector(ele, selector):
    return ele.select(selector)


def find_by_id(ele, tag, id_):
    return ele.find(tag, attrs={'id': id_})


def find_all_by_id(ele, tag, id_):
    return ele.find_all(tag, attrs={'id': id_})

# curd


def delete_ele(ele):
    ele.extract()

# img


def dict_tag_img(img):
    if 'alt' in img.attrs:
        return {'alt': img.attrs['alt'], 'href': img.attrs['href']}
    else:
        return {'alt': get_text(img), 'href': img.attrs['href']}


def img_dict_list(ele):
    return [dict_tag_img(img) for img in ele.find_all('img')]

# attr


def remove_attrs(ele):
    '''
    remove css style
    '''
    attrs = [k for k, v in ele.attrs.items()]
    for attr in attrs:
        del ele.attrs[attr]
    for ele_ in ele.find_all():
        attrs = [k for k, v in ele_.attrs.items()]
        for attr in attrs:
            del ele_.attrs[attr]
    return str(ele)

# table


def table2dict(table):
    '''
    <table>xxx</table>
    -->
    dict_table_records
    '''
    trs = table.find_all('tr')
    columns = [th.get_text() for th in (trs[0].find_all('th')
                                        if trs[0].find_all('th') else trs[0].find_all('td'))]
    records = [dict(zip(columns, [td.get_text()
                                  for td in tr.find_all('td')])) for tr in trs[1:]]
    return records


def vertical_table2dict(table):
    d = {}
    for tr in table.find_all('tr'):
        tds = tr.find_all('td')
        for ix in range(0, len(tds), 2):
            d.update({
                tds[ix].get_text().strip(): tds[ix+1].get_text().strip()
            })
    return d


def extract_embed_table(v):
    '''
    xxx<table>xxx</table>xxx
    -->
    texts:
    xxx
    table@index
    xxx
    tables:
    dict_table_records/raw_html_table
    '''
    bs = BeautifulSoup(v, 'lxml')
    if '<table>' in v:
        dicts_ = []
        for ix, table in enumerate(bs.find_all('table')):
            #dict_ = table2dict(table)
            dict_ = remove_attrs(table)
            table.replace_with('table@%s' % (str(ix)))
            dicts_.append(dict_)
        return {'texts': [i for i in bs.stripped_strings], 'tables': dicts_}
    else:
        return [i for i in bs.stripped_strings]


def table_back(d):
    if type(d) == dict:
        if 'texts' in d.keys():
            temp = []
            for text in d['texts']:
                table_index = [int(i) for i in re.findall('table@(\d+)', text)]
                if table_index:
                    table_index = table_index[0]
                    temp.append(d['tables'][table_index])
                else:
                    temp.append(text)
            return temp
        else:
            for k, v in d.items():
                d.update({k: table_back(v)})
            return d
    else:
        return d


def dict2html_table(dict_):
    data_split = pd.DataFrame.from_records(dict_).to_dict(orient='split')
    table = '<table>'
    table += '<tr>'
    for column in data_split['columns']:
        table += '<th>'+column+'</th>'
    table += '</tr>'
    for line in data_split['data']:
        if any(line):
            table += '<tr>'
            for i in line:
                table += '<td>'+str(i)+'</td>'
            table += '</tr>'
    table += '</table>'
    return table

# auto


def apply_viewer(url_or_bs_viewer_tuple):
    '''
    viewer:自定义json格式
    selector:全局 css定位局部根节点
        谨慎检查error信息,页面结构变化影响css定位的准确性
    func:局部 bs解析函数【扩展解析函数库】
        谨慎使用非本模块的函数以及被装饰器包装后的函数
    '''
    ele, viewer = url_or_bs_viewer_tuple
    res = {'url': ele} if isinstance(ele, str) else {}
    try:
        ele = BS(ele) if isinstance(ele, str) else ele
        for k, v in viewer.items():
            if v.get('selector', None):
                res.update({
                    k: v.get('func')(ele.select_one(v.get('selector')))
                })
            elif v.get('text', None):
                res.update({
                    k: v.get('func')(ele.find(text=re.compile(v.get('text'))))
                })
            elif v.get('attrs', None):
                res.update({
                    k: v.get('func')(
                        ele.find(v.get('tag'), attrs=v.get('attrs')))
                })
            else:
                res.update({
                    k: apply_viewer((ele, v))
                })
        return res
    except Exception as e:
        res.update({'error': str(e)})
        return res


def auto_local_spider(urls, viewer, path=None):
    '''
    css定位-bs抽取
    dask并行化
    '''
    sequence = [(url, viewer) for url in urls]
    res_json = dask_bag_sequence_parallel(sequence, apply_viewer)
    if path:
        dump_json(res_json, path)
    else:
        return res_json
